<?php

namespace JimsCustomerBooking\Database;

use Defuse\Crypto\Exception\BadFormatException;
use Defuse\Crypto\Exception\EnvironmentIsBrokenException;
use JimsCustomerBooking\Helpers\HashHelper;
use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\JimsPlugin;
use JimsCustomerBooking\Services\DivisionService;

/**
 * Class Booking.
 */
class Booking extends Model
{
    protected $table_name = 'customer_bookings';

    /**
     * Returns the ID of the local booking.
     *
     * @param null|mixed $error_code
     * @param mixed      $mode
     * @param mixed      $channel
     * @param mixed      $data
     * @param mixed      $status
     *
     * @return int
     *
     * @throws BadFormatException|EnvironmentIsBrokenException|\JsonException
     */
    public function add($data, $status = 'pending', $error_code = null, $mode = 'production', $channel = 'online')
    {
        $title = $this->calculateTitleFromData($data);
        $division_id = 0;

        if (self::thereIsTheSameToday(json_encode($data))) {
            return 0;
        }

        if (!empty($data['division_id'])) {
            $division_id = $data['division_id'];
        }

        return $this->insert([
            'title' => $title,
            'booking_id' => $this->getBookingHash($title),
            'division_id' => $division_id,
            //            'data' => HashHelper::create(\json_encode($data, JSON_THROW_ON_ERROR), JimsPlugin::ENCRYPTION_KEY),
            'data' => \json_encode($data, JSON_THROW_ON_ERROR),
            'status' => $status,
            'mode' => $mode,
            'channel' => $channel, // default channel is 'online' or 'manual'
            'error_code' => $error_code,
            'created_at' => gmdate('Y-m-d H:i:s'),
            'updated_at' => gmdate('Y-m-d H:i:s'),
        ]);
    }

    /**
     * @param string $output
     * @param mixed  $id
     */
    public function getFromId($id, $output = self::RESULT_TYPE_ARRAY_A): ?array
    {
        return parent::getFromId($id, $output);
    }

    // @throws \JsonException
    public function updateStatus(int $id, string $status, int $error_code): void
    {
        $table = $this->getTable();

        // get booking
        $booking = $this->getFromId($id, self::RESULT_TYPE_ARRAY_A);

        // get data and decode
        // $data = HashHelper::deHash(
        //     json_decode($booking['data'], true, 512, JSON_THROW_ON_ERROR),
        //     JimsPlugin::ENCRYPTION_KEY
        // );
        $data = json_decode($booking['data'], true, 512, JSON_THROW_ON_ERROR);

        if ($this->isTest($data)) {
            $error_code = 7;
        }

        $this->update($table, ['status' => $status, 'error_code' => $error_code], ['id' => $id]);
    }

    public function isTest($data, $indicator_keys = ['firstname', 'lastname'])
    {
        // if any of the keys exists and is equal to test return true
        foreach ($indicator_keys as $key) {
            if (!empty($data[$key]) && $data[$key] === 'test') {
                return true;
            }
        }
    }

    /**
     * @param mixed $page
     * @param mixed $limit
     * @param mixed $status
     *
     * @return null|array
     */
    public function getPage($page = 1, $limit = 25, $status = 'all')
    {
        $offset = ($page - 1) * $limit;
        $table = $this->getTable();

        if ($status == 'all') {
            $sql = "SELECT * FROM {$table} ORDER BY created_at ASC LIMIT %d OFFSET %d";
            $sql = $this->prepare($sql, [$limit, $offset]);
        } else {
            $sql = "SELECT * FROM {$table} WHERE booking_status = %s ORDER BY created_at ASC LIMIT %d OFFSET %d";
            $sql = $this->prepare($sql, [$status, $limit, $offset]);
        }

        // apply filter
        $sql = apply_filters(StringHelper::getPrefixed('query_booking_get_page'), $sql, $page, $limit, $status);

        return $this->getResults($sql, self::RESULT_TYPE_ARRAY_A);
    }

    public static function getPageCount($limit = 10, $status = 'all'): int
    {
        // Can this be new self? No class extends Booking so it can be self right?
        $self = new self();
        $table = $self->getTable();

        if ($status === 'all') {
            $sql = "SELECT COUNT(*) FROM {$table}";
        } else {
            $sql = "SELECT COUNT(*) FROM {$table} WHERE status = %s";
            $sql = $self->prepare($sql, [$status]);
        }

        // apply filter
        $sql = apply_filters(StringHelper::getPrefixed('query_booking_get_page_count'), $sql, $limit, $status);

        $all_records_count = $self->db->get_var($sql);

        // ceil is still a float. Thanks PHP
        // https://www.php.net/manual/en/function.ceil.php
        return (int) ceil($all_records_count / $limit);
    }

    public static function getPageCountForPeriod($start_date, $end_date, $per_page, $status, $timezone = '+00:00'): int
    {
        // Can this be new self? No class extends Booking so it can be self right?
        $self = new self();
        $table = $self->getTable();

        $created_at = self::getCreatedAtForTimeZone($timezone);

        if ($status === 'all') {
            // phpcs:ignore
            $sql = "SELECT COUNT(*) FROM {$table} WHERE {$created_at} BETWEEN STR_TO_DATE('{$start_date}', '%d/%m/%Y') AND STR_TO_DATE('{$end_date}', '%d/%m/%Y')";
        } else {
            // phpcs:ignore
            $sql = "SELECT COUNT(*) FROM {$table} WHERE {$created_at} BETWEEN STR_TO_DATE('{$start_date}', '%d/%m/%Y') AND STR_TO_DATE('{$end_date}', '%d/%m/%Y') AND status = '{$status}'";
            //            $sql = $self->prepare($sql, [$status]);
        }

        // apply filter
        $sql = apply_filters(StringHelper::getPrefixed('query_booking_get_page_count'), $sql, $per_page, $status);

        // ceil is still a float. Thanks PHP
        // https://www.php.net/manual/en/function.ceil.php
        return (int) ceil($self->db->get_var($sql) / $per_page);
    }

    /**
     * @param null|mixed $start_date
     * @param null|mixed $end_date
     * @param mixed      $timezone
     * @param mixed      $status
     *
     * @return null|array|object|\stdClass[]
     */
    public static function getStatisticsTotals(
        $status = 'sended',
        $start_date = null,
        $end_date = null,
        $timezone = '+00:00'
    ) {
        // Can this be new self? No class extends Booking so it can be self right?
        $self = new self();
        $table = $self->getTable();

        if (empty($start_date)) {
            // phpcs:ignore
            $sql = "SELECT title, COUNT(*) AS total_bookings, SUM(CASE WHEN status = %s THEN 1 ELSE 0 END) AS total_bookings_with_status FROM {$table} GROUP BY title ORDER BY title, created_at;";
            $sql = $self->prepare($sql, [$status]);
        } else {
            // check if start_date is formatted like a date
            if (preg_match('/^\d{2}\/\d{2}\/\d{4}$/', $start_date) === 0) {
                return [];
            }

            if (empty($end_date)) {
                $end_date = date('d/m/Y');
            }

            // get created_at field with timezone
            $created_at = self::getCreatedAtForTimeZone($timezone);

            // phpcs:ignore
            $sql = "SELECT title, COUNT(*) AS total_bookings, SUM(CASE WHEN status = '{$status}' THEN 1 ELSE 0 END) AS total_bookings_with_status FROM {$table} WHERE {$created_at} BETWEEN STR_TO_DATE('{$start_date}', '%d/%m/%Y') AND STR_TO_DATE('{$end_date}', '%d/%m/%Y') GROUP BY title ORDER BY title, created_at;";
        }

        return $self->getResults($sql, self::RESULT_TYPE_ARRAY_A);
    }

    /**
     * @param mixed      $timezone
     * @param null|mixed $date
     * @param mixed      $status
     *
     * @return null|array|object|\stdClass[]
     */
    public static function getStatisticsTotalsByDate($date = null, $status = 'sended', $timezone = '+00:00')
    {
        $self = new self();
        $table = $self->getTable();

        if ($date === null) {
            $date = date('d/m/Y');
        }

        // get created_at field with timezone
        $created_at = self::getCreatedAtForTimeZone($timezone);

        // phpcs:ignore
        $sql = "SELECT title, COUNT(*) AS total_bookings, SUM(CASE WHEN status = '{$status}' THEN 1 ELSE 0 END) AS total_bookings_with_status FROM {$table} WHERE {$created_at} = STR_TO_DATE('{$date}', '%d/%m/%Y') GROUP BY title ORDER BY title, created_at;";

        return $self->getResults($sql, self::RESULT_TYPE_ARRAY_A);
    }

    /**
     * @param mixed $offset
     * @param mixed $limit
     *
     * @return null|array|object|\stdClass[]
     */
    public static function getStatisticsAll($offset = 0, $limit = 10)
    {
        $self = new self();
        $table = $self->getTable();

        $sql = "SELECT * FROM {$table} ORDER BY created_at ASC LIMIT %d OFFSET %d";
        $sql = $self->prepare($sql, [$limit, $offset]);

        return $self->getResults($sql, self::RESULT_TYPE_ARRAY_A);
    }

    public static function getStatisticsAllPageCount(): int
    {
        $self = new self();
        $table = $self->getTable();

        $sql = "SELECT count(id) FROM {$table}";
        $result = $self->db->get_var($sql);

        return !is_null($result)
            ? (int) $result
            : 0;
    }

    /**
     * @param       $status     string
     * @param       $start_date string
     * @param       $end_date   string
     * @param       $offset     int
     * @param       $limit      int
     * @param mixed $timezone
     *
     * @return null|array|object|\stdClass[]
     */
    public static function getStatisticsAllByPeriod(
        string $status = 'sended',
        $start_date = null,
        $end_date = null,
        $offset = 0,
        $limit = 10,
        $timezone = '+00:00'
    ) {
        $self = new self();
        $table = $self->getTable();

        if ($start_date === null) {
            $start_date = date('d/m/Y');
        }

        if ($end_date === null) {
            $end_date = date('d/m/Y');
        }

        // get created_at field with timezone
        $created_at = self::getCreatedAtForTimeZone($timezone);

        // phpcs:ignore
        $sql = "SELECT * FROM {$table} WHERE {$created_at}  BETWEEN STR_TO_DATE('{$start_date}', '%d/%m/%Y') AND STR_TO_DATE('{$end_date}', '%d/%m/%Y') ORDER BY created_at ASC LIMIT {$limit} OFFSET {$offset}";

        return $self->getResults($sql, self::RESULT_TYPE_ARRAY_A);
    }

    /**
     * @param mixed $minutes
     * @param mixed $data
     *
     * @return bool
     */
    public static function thereIsTheSameToday($data, $minutes = 30)
    {
        $self = new self();
        $table = $self->getTable();

        // compare if the current data are already in the database with timespamp of max 30 minutes from now
        // phpcs:ignore
        $sql = "SELECT * FROM {$table} WHERE DATE(created_at) = CURDATE() AND data = %s and error_code in (0,1,2,3,4,8) and TIMESTAMPDIFF(MINUTE, created_at, NOW()) < %d";

        //        $sql = "SELECT * FROM $table WHERE DATE(created_at) = CURDATE() AND data = %s";
        $sql = $self->prepare($sql, [$data, $minutes]);

        return count($self->getResults($sql, self::RESULT_TYPE_ARRAY_A)) > 0;
    }

    /**
     * @param mixed      $timezone
     * @param mixed      $status
     * @param null|mixed $start_date
     * @param null|mixed $end_date
     *
     * @return mixed
     */
    public static function getStatisticsPeriodTotal(
        $status = 'sended',
        $start_date = null,
        $end_date = null,
        $timezone = '+00:00'
    ) {
        $self = new self();
        $table = $self->getTable();

        if ($start_date === null) {
            $start_date = date('d/m/Y');
        }

        if ($end_date === null) {
            $end_date = date('d/m/Y');
        }

        // get created_at field with timezone
        $created_at = self::getCreatedAtForTimeZone($timezone);

        // phpcs:ignore
        $sql = "SELECT COUNT(*) AS total_bookings, SUM(CASE WHEN status = '{$status}' THEN 1 ELSE 0 END) AS total_bookings_with_status, (SUM(CASE WHEN status = '{$status}' THEN 1 ELSE 0 END) / COUNT(*)) as rate FROM {$table} WHERE {$created_at} BETWEEN STR_TO_DATE('{$start_date}', '%d/%m/%Y') AND STR_TO_DATE('{$end_date}', '%d/%m/%Y')";

        return $self->db->get_row($sql, self::RESULT_TYPE_ARRAY_A);
    }

    public static function getCreatedAtForTimeZone($timezone): string
    {
        if (empty($timezone) || $timezone === '+00:00') {
            return 'DATE(created_at)';
        }

        return "CONVERT_TZ(created_at, '+00:00', '{$timezone}')";
    }

    public function getBookingHash(string $title): string
    {
        return md5(time() . $title);
    }

    /**
     * @param mixed $data
     *
     * @throws \Exception
     */
    protected function calculateTitleFromData($data)
    {
        // get country code
        $country_code = 'AU';
        if (!empty($data['address']['country'])) {
            $country_code = $data['address']['country'];
        }

        // get services from FMS
        $divisionService = new DivisionService();
        $services = $divisionService->getServices(true, $country_code);

        // get first service id
        $first_service = $data['services'][0];

        // set default name the first service id
        $name = $first_service;

        // get the name of the first service
        if (!empty($services) && is_array($services)) {
            $service = array_filter($services, static function ($item) use ($first_service) {
                return (int) $item['id'] === (int) $first_service;
            });

            $service = array_shift($service);

            $name = !empty($service['name']) && is_string($service['name']) ? $service['name'] : $name;
        }

        return $name;
    }
}
