<?php

namespace JimsCustomerBooking\Database;

use JimsCustomerBooking\JimsPlugin;

/**
 * Class Model.
 *
 * Database Model
 * use the wpdb class to interact with the database
 */
abstract class Model
{
    public const RESULT_TYPE_DEFAULT = ARRAY_N;
    public const RESULT_TYPE_OBJECT = OBJECT;
    public const RESULT_TYPE_ARRAY_A = ARRAY_A;
    public const RESULT_TYPE_ARRAY_N = ARRAY_N;
    public const RESULT_TYPE_OBJECT_K = OBJECT_K;

    /**
     * @var \wpdb
     */
    protected $db;

    protected $table_name = '';

    protected $results_limit = 25;

    /**
     * Model constructor.
     */
    public function __construct()
    {
        global $wpdb;
        $this->db = $wpdb;
    }

    public function __call($name, $arguments)
    {
        if (method_exists($this->db, $name)) {
            return call_user_func_array([$this->db, $name], $arguments);
        }
    }

    public function getTable(): string
    {
        return $this->db->prefix . implode('_', [JimsPlugin::PREFIX, $this->table_name]);
    }

    public function prepare(string $sql, mixed $args = []): string
    {
        return $this->db->prepare($sql, $args);
    }

    /**
     * @return null|bool|int|\mysqli_result|resource
     */
    public function query(string $sql, mixed $args = [])
    {
        return $this->db->query(
            $this->db->prepare($sql, $args)
        );
    }

    public function update($table, $data, $where, $format = null, $where_format = null): void
    {
        $this->db->update($table, $data, $where, $format, $where_format);
    }

    /**
     * @param mixed $data
     *
     * @return int
     */
    public function insert($data)
    {
        $table = $this->getTable();
        $this->db->insert($table, $data);

        return $this->db->insert_id;
    }

    public function select(
        $fields = '*',
        $where = [],
        $operator = 'AND',
        $limit = 25,
        $offset = 0,
        $order_by = 'id',
        $order = 'DESC'
    ) {
        $table = $this->getTable();
        $fields = $fields != '*' ? implode(',', $fields) : $fields;

        list($where, $directive_string) = $this->getSelectConditions(
            $where,
            $operator,
            $limit,
            $offset,
            $order_by,
            $order
        );

        return $this->getResults("SELECT {$fields} FROM {$table}" . $where . $directive_string);
    }

    /**
     * @param mixed $output
     * @param mixed $sql
     *
     * @return null|array|object|\stdClass[]
     */
    public function getResults($sql, $output = self::RESULT_TYPE_ARRAY_A)
    {
        return $this->db->get_results($sql, $output);
    }

    public function getCount($where = [], $operator = 'AND')
    {
        $table = $this->getTable();
        list($where, $directive_string) = $this->getSelectConditions($where, $operator, 0, 0, 'id', 'DESC');
        $sql = "SELECT COUNT(*) FROM {$table}" . $where;

        return $this->db->get_var($sql);
    }

    /**
     * @param mixed $output
     * @param mixed $id
     *
     * @return null|array|object|\stdClass
     */
    public function getFromId($id, $output = self::RESULT_TYPE_OBJECT)
    {
        $table = $this->getTable();
        $sql = "SELECT * FROM {$table} WHERE id = {$id}";

        return $this->db->get_row($sql, $output);
    }

    public function getDateField($field_name)
    {
        $this->getAsDatetime($field_name);
    }

    /**
     * @param mixed $timestamp
     *
     * @return string
     */
    protected function getAsDatetime($timestamp)
    {
        return date('Y-m-d H:i:s', $timestamp);
    }

    protected function getSelectConditions(
        mixed $where,
        mixed $operator,
        mixed $limit,
        mixed $offset,
        mixed $order_by,
        mixed $order
    ): array {
        if (!empty($where) && is_array($where)) {
            $where = sprintf(' WHERE %s', implode(" {$operator} ", $where));
        } else {
            $where = '';
        }

        if ($limit == 0 || empty($limit)) {
            $limit = $this->results_limit;
        }

        $directive_string = sprintf(' LIMIT %d', $limit);

        if ($offset > 0) {
            $directive_string .= sprintf(' OFFSET %d', $offset);
        }

        $directive_string = sprintf(' ORDER BY %s %s', $order_by, $order);

        return [$where, $directive_string];
    }
}
