<?php

namespace JimsCustomerBooking\Helpers;

class FSHelper
{
    /**
     * Returns an array of pathnames matching a pattern within the plugin directory.
     *
     * @param $path     See https://www.php.net/manual/en/function.glob.php
     * @param $flags    See https://www.php.net/manual/en/function.glob.php
     * @param $relative Whether or not the pathnames returned are relative to the plugin directory
     */
    final public static function glob(string $path, int $flags = 0, bool $relative = false): array
    {
        $pathnames = glob(static::path($path), $flags);

        if (!is_array($pathnames)) {
            return [];
        }

        return $relative === true
            ? array_map(function ($file) {
                return basename($file);
            }, $pathnames)
            : $pathnames;
    }

    /**
     * Returns a json decoded string from a file within the plugin. By default,
     * an obect is returned.
     *
     * @param $file    A file relative to the plugin directory containing a json encoded string
     * @param $asArray If true, the json will be decoded to an array
     */
    final public static function json(string $file, bool $asArray = false): null|array|object
    {
        if (empty(pathinfo($file, PATHINFO_EXTENSION))) {
            $file .= '.json';
        }

        return json_decode(static::content($file), $asArray);
    }

    /**
     * Return the content of a file located within the plugin.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function content(string $file): string
    {
        return \trim(\file_get_contents(
            static::path($file)
        ));
    }

    /**
     * Returns whether or not a path exists within the plugin, and if that path
     * represents a file.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function fileExists(string $file): bool
    {
        $path = static::path($file);

        return file_exists($path) && is_file($path);
    }

    /**
     * Returns whether or not a path exists within the plugin.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function exists(string $file): bool
    {
        return \file_exists(
            static::path($file)
        );
    }

    /**
     * Requires a file and returns the contents.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function require(string $file): mixed
    {
        return require static::path($file);
    }

    /**
     * A URL to a file/asset within the plugin.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function url(string $file = ''): string
    {
        return trailingslashit(JIMS_CBP_URL) . $file;
    }

    /**
     * A URL to a file/asset within the plugin.
     *
     * @param $file A file relative to the plugin directory
     */
    final public static function path(string $file = ''): string
    {
        return trailingslashit(JIMS_CBP_DIR) . $file;
    }
}
