<?php

namespace JimsCustomerBooking\Helpers;

use JimsGroup\Utilities\Logger;

if (!class_exists('JimsCustomerBooking\Helpers\Http')) {
    class Http extends Helper
    {
        public static array $curl_options = [];

        /**
         * @param mixed $use_http_compatibility_mode
         *
         * @return array
         */
        public static function get(string $url, array $params = [], $use_http_compatibility_mode = false)
        {
            if ($use_http_compatibility_mode) {
                self::setHttpCompatibilityMode();
            }

            return self::request($url, 'GET', $params);
        }

        public static function setHttpCompatibilityMode(string $version = '1.1'): void
        {
            $versions = [
                'none' => CURL_HTTP_VERSION_NONE, // let curl decide
                '1.0' => CURL_HTTP_VERSION_1_0,
                '1.1' => CURL_HTTP_VERSION_1_1,
                '2.0' => CURL_HTTP_VERSION_2_0,
            ];

            if (isset($versions[$version])) {
                self::setCurlOptions([
                    CURLOPT_HTTP_VERSION => $versions[$version],
                ]);
            } else {
                self::setCurlOptions([
                    CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
                ]);
            }
        }

        /**
         * @param mixed $use_http_compatibility_mode
         *
         * @return array
         */
        public static function post(
            string $url,
            array $params = [],
            array $extra_headers = [],
            $use_http_compatibility_mode = false
        ) {
            Logger::log($use_http_compatibility_mode ? 'true' : 'false', 'info, post params');

            if ($use_http_compatibility_mode) {
                self::setHttpCompatibilityMode();
            }

            return self::request($url, 'POST', $params, $extra_headers);
        }

        /**
         * @param mixed $timeout
         *
         * @return array
         */
        public static function request(
            string $url,
            string $verb = 'GET',
            array $params = [],
            array $extra_headers = [],
            $timeout = false
        ) {
            // use curl to make the request
            $connection = curl_init();

            $headers = self::getHeadersForContentType($extra_headers['Content-Type'] ?? 'application/json');

            if (empty($timeout)) {
                $timeout = 0;
            }

            if ($verb === 'GET' && count($params) > 0) {
                $url .= '?' . http_build_query($params);
            }

            $default_options = [
                CURLOPT_URL => $url,
                CURLOPT_RETURNTRANSFER => true,
                CURLOPT_ENCODING => '',
                CURLOPT_TIMEOUT => $timeout,
                CURLOPT_HTTPHEADER => array_merge($headers, $extra_headers),
            ];

            if ($verb !== 'GET' && $verb !== 'HEAD') {
                $default_options[CURLOPT_CUSTOMREQUEST] = $verb;
                $default_options[CURLOPT_POSTFIELDS] = self::getBodyForContentType(
                    $extra_headers['Content-Type'] ?? 'application/json',
                    $params
                );
            }

            // log options before combining
            Logger::log($default_options, 'info, default curl options');

            $options = $default_options;
            if (in_array(CURLOPT_HTTP_VERSION, self::$curl_options)) {
                $options[CURLOPT_HTTP_VERSION] = self::$curl_options[CURLOPT_HTTP_VERSION];
            }

            // log options after combining
            Logger::log($options, 'info, curl options');

            // set the url, number of POST vars, POST data with opt array
            curl_setopt_array($connection, $options);

            // execute post
            $result = curl_exec($connection);

            return [
                'result' => $result,
                'headers' => curl_getinfo($connection),
                'error' => curl_error($connection),
            ];
        }

        /**
         * This method tests if the API is available.
         *
         * @param mixed $url
         *
         * @return bool
         */
        public static function makeTestRequest($url)
        {
            $result = self::request($url, 'GET', [], [], 10);

            return $result['headers']['http_code'] == 200;
        }

        /**
         * @param mixed $content_type
         *
         * @return string[]
         */
        public static function getHeadersForContentType($content_type)
        {
            switch ($content_type) {
                case 'application/json':
                    return [
                        'Content-Type: application/json',
                        'Accept: application/json',
                    ];

                case 'application/x-www-form-urlencoded':
                    return [
                        'Content-Type: application/x-www-form-urlencoded',
                        'Accept: application/json',
                    ];

                default:
                    return [
                        'Content-Type: application/json',
                        'Accept: application/json',
                    ];
            }
        }

        /**
         * @param mixed $content_type
         * @param mixed $params
         *
         * @return string
         */
        public static function getBodyForContentType($content_type, $params)
        {
            switch ($content_type) {
                case 'application/json':
                    return \json_encode($params);

                case 'application/x-www-form-urlencoded':
                    return \http_build_query($params);

                default:
                    return \json_encode($params);
            }
        }

        public static function getUserIP()
        {
            if (!empty($_SERVER['HTTP_CLIENT_IP'])) {
                $ip = $_SERVER['HTTP_CLIENT_IP'];
            } elseif (!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
                $ip = $_SERVER['HTTP_X_FORWARDED_FOR'];
            } else {
                $ip = $_SERVER['REMOTE_ADDR'];
            }

            return $ip;
        }

        public static function getCurlOptions(): array
        {
            return self::$curl_options;
        }

        public static function setCurlOptions(array $curl_options, bool $reset = false): array
        {
            if (count($curl_options) > 0) {
                $options = $reset || count(self::$curl_options) === 0
                    ? $curl_options
                    : array_merge(self::$curl_options, $curl_options);
                self::$curl_options = $options;
            }

            return self::$curl_options;
        }

        public static function combineCurlOptions(array $default_options): array
        {
            $options = self::$curl_options;

            foreach ($options as $key => $value) {
                if (isset($default_options[$key]) || $key === CURLOPT_HTTPHEADER) {
                    continue;
                }

                $default_options[(int) $key] = $value;
            }

            if (isset($options[CURLOPT_HTTPHEADER])) {
                $default_options[CURLOPT_HTTPHEADER] = array_merge(
                    $default_options[CURLOPT_HTTPHEADER],
                    $options[CURLOPT_HTTPHEADER]
                );
            }

            return $default_options;
        }
    }
}
