<?php

namespace JimsCustomerBooking\Helpers;

use JimsCustomerBooking\Admin\BackendPages\NativeFormSettingsPage;
use JimsCustomerBooking\JimsOptions;
use JimsCustomerBooking\JimsPlugin;
use JimsCustomerBooking\Services\TokenManager;

if (!class_exists('JimsCustomerBooking\Helpers\OptionHelper')) {
    class OptionHelper extends Helper
    {
        /**
         * @param mixed      $option_name
         * @param null|mixed $default
         *
         * @return false|mixed|void
         */
        public static function get($option_name, $default = null)
        {
            return get_option(StringHelper::getPrefixed($option_name, 'machine'), $default);
        }

        public static function update($option_name, $value): bool
        {
            return update_option(StringHelper::getPrefixed($option_name, 'machine'), $value);
        }

        public static function getValue($option_name, $value_name, $default = null)
        {
            $option = self::get($option_name, []);

            return $option[$value_name] ?? $default;
        }

        /**
         * @return JimsOptions
         */
        public static function getEmbeddedFormOptions()
        {
            $default = new JimsOptions();
            $options = self::get('form', $default);

            if ($options instanceof JimsOptions) {
                return $options;
            }

            if (is_string($options)) {
                $options = \unserialize($options);
            }

            return $options instanceof JimsOptions ? $options : $default;
        }

        public static function localizeFormVariables($script_name)
        {
            $settings = self::getGlobalSettings('admin');

            wp_add_inline_script(
                StringHelper::getPrefixed($script_name),
                'window.' . StringHelper::getPrefixed('variables', 'machine') . ' = ' . json_encode($settings),
                'before'
            );
        }

        public static function getGlobalSettings($context = 'frontend'): array
        {
            $settings = static::get(NativeFormSettingsPage::OPTION_NAME, []);

            // remove options with key starting with 'required_fields'
            $settings = array_filter($settings, function ($key) {
                return !str_contains($key, 'required_fields');
            }, ARRAY_FILTER_USE_KEY);

            // getRequiredInfoFields returns an array with non sequential indexes.
            // We only need the values for rendering the form
            $settings['requiredFields'] = array_values(self::getRequiredInfoFields());

            $finalSettings = array_merge(
                [
                    'home_url' => get_bloginfo('wpurl'),
                    'version' => JimsPlugin::VERSION,
                    'rest_url' => get_rest_url(),
                    'secret' => wp_create_nonce('wp_rest'),
                    'session_token' => TokenManager::encode(
                        JimsPlugin::SECKEY,
                        JimsPlugin::PREFIX . TokenManager::generateRandomString(10),
                        '10 minutes'
                    ),
                ],
                $settings,
                ['embeddedFormOptions' => JimsPlugin::getOptions()->getProperties()],
            );

            return $context === 'frontend'
                ? self::normalizeVariables($finalSettings)
                : $finalSettings;
        }

        private static function normalizeVariables($variables)
        {
            $allowedVariables = [
                'bookingFormDivisionId', 'country', 'google_api_key',
                'home_url', 'redirect_success_url', 'requiredFields',
                'use_google_autocomplete', 'use_redirect',
            ];

            $booleans = ['use_google_autocomplete', 'use_redirect'];

            // Strip keys that are not needed in the frontend
            foreach ($variables as $key => $value) {
                if (!in_array($key, $allowedVariables)) {
                    unset($variables[$key]);
                }
            }

            // Convert strings to booleans
            foreach ($booleans as $key) {
                $variables[$key] = boolval($variables[$key]);
            }

            return $variables;
        }

        public static function getArrayFirstValue(array $arr)
        {
            return array_values($arr)[0] ?? null;
        }

        public static function getFormOption($name, $default = '')
        {
            $options = self::get(NativeFormSettingsPage::OPTION_NAME, []);

            return $options[$name] ?? $default;
        }

        /**
         * @param string $target, available options are 'Internal', 'FMS'
         */
        public static function isHTTPCompatibilityEnabled(string $target = 'FMS'): bool
        {
            if (in_array($target, ['Internal', 'FMS'])) {
                return (int) self::getFormOption('httpVersionFor' . $target, 0) === 1;
            }

            return false;
        }

        public static function getRequiredInfoFields()
        {
            $options = OptionHelper::get(NativeFormSettingsPage::OPTION_NAME, []);

            $fields = ['address'];

            foreach ($options as $key => $value) {
                if (str_starts_with($key, 'required_fields-')) {
                    $fields[] = $value;
                }
            }

            if (count($fields) === 1) {
                $fields[] = 'firstname';
                $fields[] = 'lastname';
                $fields[] = 'email';
                $fields[] = 'phone';
            }

            return $fields;
        }
    }
}
