<?php

namespace JimsCustomerBooking\Helpers;

use JimsCustomerBooking\JimsPlugin;
use JimsCustomerBooking\Validators\JsonValidator;

if (!class_exists('JimsCustomerBooking\Helpers\StringHelper')) {
    class StringHelper extends Helper
    {
        public const PREFIXED_TYPE_SLUG = 'slug';
        public const PREFIXED_TYPE_MACHINE = 'machine';

        /**
         * @param string $string
         * @param string $type   , slug|machine
         *
         * @return string
         */
        public static function getPrefixed($string, $type = self::PREFIXED_TYPE_SLUG)
        {
            list($prefix, $replacement) = self::getReplacementValues($type);

            $result = preg_replace('#[\-_ ]#', $replacement, "{$prefix} {$string}");

            return is_string($result) ? strtolower($result) : strtolower(implode($replacement, $result));
        }

        public static function getPrefix($type = self::PREFIXED_TYPE_SLUG)
        {
            list($prefix, $replacement) = self::getReplacementValues($type);

            return preg_replace('#[\-_ ]#', $replacement, "{$prefix}");
        }

        /**
         * Same as getPrefixed but takes an array of parts.
         *
         * @param array $parts
         * @param mixed $type
         *
         * @return string
         */
        public static function getPrefixedFromParts($parts, $type = self::PREFIXED_TYPE_SLUG)
        {
            $string = implode(' ', $parts);

            return self::getPrefixed($string, $type);
        }

        /**
         * @param mixed $string
         *
         * @return bool
         */
        public static function isJson($string)
        {
            return JsonValidator::isJson($string);
        }

        /**
         * @param mixed $string
         *
         * @return bool
         */
        public static function isSerialized($string)
        {
            return @unserialize($string) !== false;
        }

        /**
         * @param $version1 string, format 1.0.0, follows the regex pattern \d+(\.\d+){0,2}
         * @param $version2 string, format 1.0.0, follows the regex pattern \d+(\.\d+){0,2}
         *
         * @return bool, true if version1 is greater than version2
         */
        public static function compareVersionStrings($version1, $version2)
        {
            // before version compare we need to format the version strings, in case they are not formatted
            $version1 = explode('.', $version1);
            $version2 = explode('.', $version2);

            // convert every version part to an integer
            $version1 = array_map('intval', $version1);
            $version2 = array_map('intval', $version2);

            // pad the version arrays with 0s
            $version1 = array_pad($version1, 3, 0);
            $version2 = array_pad($version2, 3, 0);

            // recompose the version strings
            $version1 = implode('.', $version1);
            $version2 = implode('.', $version2);

            // compare the versions
            return version_compare($version1, $version2, '>');
        }

        public static function getNewZealandSpecialCharacters()
        {
            return [
                'ā' => 'a',
                'ē' => 'e',
                'ī' => 'i',
                'ō' => 'o',
                'ū' => 'u',
                'Ā' => 'A',
                'Ē' => 'E',
                'Ī' => 'I',
                'Ō' => 'O',
                'Ū' => 'U',
                '-' => '-',
            ];
        }

        /**
         * @param mixed $string
         *
         * @return string
         */
        public static function convertSpecialNZSpecialChars($string)
        {
            return strtr($string, self::getNewZealandSpecialCharacters());
        }

        /**
         * @param mixed $array
         *
         * @return array
         */
        public static function convertSpecialNZCharactersInArrayValues($array)
        {
            $result_array = [];
            foreach ($array as $key => $value) {
                $result_array[$key] = self::convertSpecialNZSpecialChars($value);
            }

            return $result_array;
        }

        public static function slugify($string)
        {
            $string = strtolower($string);
            $string = preg_replace('/[^a-z0-9 -]+/', '', $string);
            $string = preg_replace('/[\s-]+/', '-', $string);

            return trim($string, '-');
        }

        /**
         * @param mixed $type
         *
         * @return array
         */
        private static function getReplacementValues($type)
        {
            return [
                JimsPlugin::PREFIX,
                [self::PREFIXED_TYPE_SLUG => '-', self::PREFIXED_TYPE_MACHINE => '_'][$type],
            ];
        }
    }
}
