<?php

namespace JimsCustomerBooking\Helpers;

/**
 * This class offers helper methods for time.
 */
class TimeHelper extends Helper
{
    /**
     * This function converts a date to UTC.
     *
     * @param string $date
     * @param string $timezone , default is UTC, other options are 'Europe/London', Melbourne/Australia
     *
     * @return \DateTime|string
     */
    public static function convertToUTC(
        $date,
        string $timezone = 'UTC',
        string $format_from = 'd/m/Y',
        string $format_to = 'd/m/Y',
        string $returnType = 'string'
    ) {
        if (empty($date)) {
            return $date;
        }

        if (is_numeric($date)) {
            $date = date($format_from, $date);
        }

        if (!self::isValidTimeZone($timezone)) {
            $timezone = 'UTC';
        }

        try {
            $date = \DateTime::createFromFormat($format_from, $date, new \DateTimeZone($timezone));
            // Set the time to 00:00:00 in the local timezone
            $date->setTime(0, 0, 0);

            // get the offset in seconds between UTC and the local timezone
            $offset = $date->getOffset();

            // Remove this offset from the date
            $date->setTimezone(new \DateTimeZone('UTC'));

            $date->add(\DateInterval::createFromDateString($offset . 'hours'));

            // Add the offset back into the date object
            return $returnType == 'string'
                ? $date->format($format_to)
                : $date;
        } catch (\Exception $e) {
            error_log($e->getMessage());
        }

        return $date;
    }

    /**
     * @param mixed $date
     *
     * @return \DateTime|string
     */
    public static function convertUTCToLocal(
        $date,
        string $timezone,
        string $format_from = 'd/m/Y H:i:s',
        string $format_to = 'd/m/Y H:i:s',
        string $returnType = 'string'
    ) {
        if (empty($date)) {
            return $date;
        }

        if (is_numeric($date)) {
            $date = date($format_from, $date);
        }

        if (!self::isValidTimeZone($timezone)) {
            $timezone = 'UTC';
        }

        try {
            $date = \DateTime::createFromFormat($format_from, $date, new \DateTimeZone('UTC'));

            // get the offset in seconds between UTC and the local timezone
            $offset = $date->getOffset();

            // Remove this offset from the date
            $date->setTimezone(new \DateTimeZone($timezone));

            // Add the offset back into the date object
            return $returnType == 'string' ? $date->format($format_to) : $date;
        } catch (\Exception $e) {
            error_log($e->getMessage());
        }

        return $date;
    }

    /**
     * @param string $timezone
     *
     * @return string
     */
    public static function getTimezoneOffset($timezone = 'UTC')
    {
        $date = new \DateTime('now', new \DateTimeZone($timezone));

        return $date->format('P');
    }

    public static function getTimezoneFromCountryCode(string $countryCode): string
    {
        switch ($countryCode) {
            case 'AU':
                return 'Australia/Melbourne';

            case 'NZ':
                return 'Pacific/Auckland';

            case 'CA':
                return 'America/Toronto';

            case 'GB':
                return 'Europe/London';

            case 'US':
                return 'America/New_York';

            default:
                return 'UTC';
        }
    }

    /**
     * Returns the timezone as UTC+{{offset}}.
     *
     * @param mixed $timezone_name
     *
     * @return string
     */
    public static function getTimezoneOffsetDBString($timezone_name)
    {
        $timezone = new \DateTimeZone($timezone_name);
        $offset = $timezone->getOffset(new \DateTime('now', $timezone));
        $offset = $offset / 60 / 60;

        return ($offset > 0 ? '+' : '') . $offset . ':00';
    }

    public static function convertDateToTimezoneFromCountryCode($date, string $countryCode): string
    {
        return self::convertToUTC(
            $date,
            self::getTimezoneFromCountryCode($countryCode)
        );
    }

    public static function isValidTimeZone(string $timezone): bool
    {
        try {
            new \DateTimeZone($timezone);

            return true;
        } catch (\Exception $e) {
            return false;
        }
    }
}
