<?php

namespace JimsCustomerBooking\Integrations;

use JimsCustomerBooking\BookingForm;
use JimsCustomerBooking\Helpers\Http;
use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\JimsPlugin;
use JimsGroup\Apis\CustomerBooking;
use JimsGroup\Datatypes\AddressType;
use JimsGroup\Datatypes\BookingResponseType;
use JimsGroup\Datatypes\Collection;
use JimsGroup\Utilities\LocationHelper;
use mysql_xdevapi\Exception;

/**
 * Class JimsFMS.
 */
class JimsFmsBooking
{
    public array $errors = [];
    private ?CustomerBooking $customer_booking = null;

    /**
     * It selects the server url
     * 'production', 'uat', 'hf', 'staging' 'hf2', 'development'.
     */
    private string $mode = 'production';

    private array $responses = [];
    private array $messages = [];
    private ?BookingForm $booking_form = null;
    private string $source_website_url = '';
    private string $api_key = JimsPlugin::FMS_API_KEY;
    private string $auth_header_token = '';

    public function __construct(
        $countryCode = 'AU',
        $auth_header_token = '',
        $mode = 'production',
        $source_website_url = ''
    ) {
        // set mode
        $this->mode = $mode;

        // initialize customer booking api
        $this->customer_booking = new CustomerBooking();

        // set mode and api key, it selects the server url
        $this->customer_booking->setMode($this->mode);

        // set api key
        $this->customer_booking->setAuthToken($this->api_key);

        $this->setAuthHeaderToken($auth_header_token);

        $this->setSourceWebsiteUrl($source_website_url);

        try {
            // get initial booking request
            $this->doInitialRequest($this->validateCountryCode($countryCode), $source_website_url);
        } catch (\Exception $e) {
            $this->errors[] = $e->getMessage();
        }
    }

    public function getApiUrl(): string
    {
        return $this->customer_booking->getApiUrl();
    }

    /**
     * @param mixed $website_home_url
     *
     * @throws \Exception
     */
    public function doInitialRequest(string $countryCode = 'AU', $website_home_url = ''): mixed
    {
        $this->checkForAuthHeaderToken();

        // get transient name
        $transientName = StringHelper::getPrefixedFromParts(['fms_initial_request', $countryCode]);

        $requestResponse = get_transient($transientName);

        // get transient
        if ($requestResponse === false) {
            $requestResponse = [];

            try {
                // get initial booking request
                $requestResponse = $this->customer_booking->getInitialBookingRequest([
                    'countryCode' => $countryCode,
                    'utmSource' => $website_home_url,
                ]);

                // set transient
                set_transient($transientName, $requestResponse, 1 * HOUR_IN_SECONDS);
            } catch (\Exception $e) {
                throw new \Exception($e->getMessage());
            }
        }

        return $requestResponse;
    }

    /**
     * @param mixed $customer
     * @param mixed $address
     * @param mixed $division_id
     *
     * @throws \Exception
     */
    public function doBookingRequest($customer, $address, $division_id, array $service_ids): bool
    {
        $this->checkForAuthHeaderToken();

        // Store the data to a BookingForm object
        $this->booking_form = new BookingForm($customer, $address, $division_id, $service_ids);

        error_log(json_encode($this->booking_form->toArray(), JSON_THROW_ON_ERROR), 0);

        // try direct booking or allocation
        $booking = false;

        try {
            $booking = $this->tryToBook($this->booking_form);
        } catch (Exception $e) {
            error_log(implode(' ', [$e->getFile(), $e->getLine(), $e->getCode(), $e->getMessage()]), 0);
        }

        return $booking;
    }

    /**
     * @param mixed $methods
     *
     * @throws \JsonException
     */
    public function tryToBook(BookingForm $form, $methods = ['doAllocationRequest', 'doFinalRequest']): bool
    {
        foreach ($methods as $method) {
            /**
             * @var BookingResponseType $response
             */
            $response = $this->{$method}($form);
            $this->addResponse($response);

            error_log(__FILE__ . ' ' . json_encode($response->toArray(), JSON_THROW_ON_ERROR), 0);

            // add check for message and status codes here
            $this->setMessage(json_encode($response->toArray(), JSON_THROW_ON_ERROR));

            $status = json_encode(
                [
                    'message' => $response->getStatusMessage(),
                    'code' => $response->getConfirmCodes(),
                ],
                JSON_THROW_ON_ERROR
            );

            if ($response->isBookingFailed()) {
                $this->setMessage($status);

                throw new \Exception($response->getStatusMessage());
            }
            if ($response->isBookingCompleted()) {
                $this->setMessage($status);

                return true;
            }
        }

        return false;
    }

    /**
     * @throws \Exception
     */
    public function doAllocationRequest(BookingForm $form): BookingResponseType
    {
        $this->checkForAuthHeaderToken();

        $address = new AddressType($form->getAddress());

        $this->customer_booking->setUtmSource($this->getSourceWebsiteUrl());

        return $this->customer_booking->sendAllocationRequest(
            $form->getDivisionId(),
            $form->getServices(),
            $address->toArray(),
            $form->getCustomerInfo(),
            $this->customer_booking->getAuthToken(),
        );
    }

    /**
     * @param null|BookingResponseType $previous_response
     * @param bool                     $used_jims_before
     * @param bool                     $is_contractor
     *
     * @throws \Exception
     */
    public function doFinalRequest(
        BookingForm $form,
        $previous_response = null,
        $used_jims_before = false,
        $is_contractor = false
    ): BookingResponseType {
        $this->checkForAuthHeaderToken();

        if ($previous_response === null) {
            $response = $this->getLastResponse();

            if ($response->result === false) {
                return new BookingResponseType(json_encode([
                    'result' => false,
                    'message' => $response->message,
                    'status' => 'failed',
                    'confirmCodes' => $response->confirmCodes,
                ], JSON_THROW_ON_ERROR));
            }
        } else {
            $response = $previous_response;
        }

        // get properties from the last response, that is not standard named
        [$booking_request_id, $provider, $provider_id] = $this->getUnsafeProperties($response);

        error_log('Used Jims Before: ' . $used_jims_before, 0);

        // get the last response and send the final request
        $params = [
            'booking_request_id' => $booking_request_id,
            'division_id' => $form->getDivisionId(),
            'service_ids' => $form->getServices(),
            'choosedProviderId' => $provider_id,
            'user_ip_address' => Http::getUserIP(),
            'provider_info' => $provider,
            'user_info' => $form->getCustomerInfo(),
            'bookingConfirmCode' => $response->getPayloadProperty('confirmCodes'),
            'previous_lead_id' => $used_jims_before ? $response->getPreviousLeadId() : 0,
            'is_contractor' => $is_contractor,
        ];

        error_log(json_encode($params, JSON_THROW_ON_ERROR), 0);

        $this->customer_booking->setUtmSource($this->getSourceWebsiteUrl());

        // phpcs:disable
        // $booking_request_id, $division_id, $service_ids, $choosedProviderId, $user_ip_address, $provider_info, $user_info, $bookingConfirmCode
        return $this->customer_booking->sendFinalizeBookingRequest(...$params);
    }

    /**
     * @param mixed $division_id
     *
     * @return Collection|mixed
     */
    public function getDivisionServices($division_id): mixed
    {
        $transientName = StringHelper::getPrefixed('fms_division_services_' . $division_id);

        if (false === ($requestResponse = get_transient($transientName))) {
            $requestResponse = $this->customer_booking->getDivisionServices($division_id);

            // store division services in transient for 12 hours
            set_transient($transientName, $requestResponse, 12 * HOUR_IN_SECONDS);
        }

        return $requestResponse;
    }

    /**
     * @return array|Collection|mixed
     */
    public function getDivisions(): mixed
    {
        $transientName = StringHelper::getPrefixed('fms_divisions');
        if (false === ($requestResponse = get_transient($transientName))) {
            $requestResponse = $this->customer_booking->getDivisions();

            // store divisions in transient for 12 hours
            set_transient($transientName, $requestResponse, 12 * HOUR_IN_SECONDS);
        }

        return $requestResponse;
    }

    public function getMessages(): array
    {
        return $this->messages;
    }

    public function setMessage($message): void
    {
        $this->messages[] = $message;
    }

    public function getResponses(): array
    {
        return $this->responses;
    }

    public function getLastResponse(): BookingResponseType
    {
        return end($this->responses);
    }

    public function getAuthHeaderToken(): string
    {
        return $this->auth_header_token;
    }

    /**
     * If the auth header token is not empty, it sets the auth header token.
     */
    public function setAuthHeaderToken(string $auth_header_token): void
    {
        if (!empty($auth_header_token)) {
            $this->auth_header_token = $auth_header_token;
            $this->customer_booking->setHeaderAuthToken($auth_header_token);
        }
    }

    public function hasHeaderAuthToken(): bool
    {
        return !empty($this->auth_header_token);
    }

    /**
     * @throws \Exception
     */
    public function checkForAuthHeaderToken(): void
    {
        if (!$this->hasHeaderAuthToken()) {
            throw new \Exception('Auth header token is not set');
        }
    }

    public function getSourceWebsiteUrl(): string
    {
        return $this->source_website_url;
    }

    public function setSourceWebsiteUrl(string $source_website_url): void
    {
        $this->source_website_url = $source_website_url;
    }

    protected function getUnsafeProperties(BookingResponseType $last_response): array
    {
        // get bookingRequestId or booking_request_id
        $booking_request_id = $last_response->getPayloadProperty('booking_request_id');
        if ($booking_request_id === null) {
            $booking_request_id = $last_response->getPayloadProperty('bookingRequestId');
        }

        // get first available fse
        $avaliableFSEs = $last_response->getPayloadProperty('avaliableFSEs');

        // set provider to null prior to checking, to allow empty provider to be returned
        // this case helps when the booking is not allocated to any provider
        // todo check if this is the best approach
        //        $provider = null;
        //        if(!empty($avaliableFSEs[0])){
        //            $provider = $avaliableFSEs[0];
        //        }

        $provider = $avaliableFSEs[0];

        // get ProviderId
        $provider_id = $last_response->getPayloadProperty('provider_id');
        if (empty($provider_id)) {
            $provider_id = $provider['business_id'];
        }

        return [$booking_request_id, $provider, $provider_id];
    }

    private function addResponse($request): void
    {
        $this->responses[] = $request;
    }

    private function validateCountryCode(string $countryCode, string $default = 'AU'): string
    {
        if (strlen($countryCode) !== 2 && !empty($countryCode)) {
            $countryCode = LocationHelper::getCountryCode($countryCode);
        }

        return $countryCode !== false
            ? $countryCode
            : $default;
    }
}
