<?php

namespace JimsCustomerBooking\Integrations;

use JimsCustomerBooking\Database\Booking;
use JimsCustomerBooking\Helpers\ArrayHelper;
use JimsCustomerBooking\Helpers\Http;
use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Helpers\StringHelper;

/**
 * Class JimsLeadsIntegration.
 */
class JimsLeadsIntegration
{
    public static ?JimsLeadsIntegration $instance = null;
    protected $servers = [
        'production' => 'https://api.leads.jims.net',
        'staging' => 'https://api-staging.leads.jims.net',
        'development' => 'http://localhost:8000',
    ];

    protected $mode = 'production';

    protected $allowed_modes = ['production', 'staging', 'development'];

    private $endpoints = [
        'staging' => [
            'update_site' => '/api/v1/instance/update',
            'update_realtime' => '/api/v1/realtime/lead',
            'email_failed' => '/api/v1/email/failed',
            'status' => '/api/v1/status',
        ],
        'production' => [
            'update_site' => '/api/v1/instance/update',
            'update_realtime' => '/api/v1/realtime/lead',
            'email_failed' => '/api/v1/email/failed',
            'status' => '/api/v1/status',
        ],
        'development' => [
            'update_site' => '/api/v1/instance/update',
            'update_realtime' => '/api/v1/realtime/lead',
            'email_failed' => '/api/v1/email/failed',
            'status' => '/api/v1/status',
        ],
    ];

    public static function getInstance(): JimsLeadsIntegration
    {
        if (self::$instance === null) {
            self::$instance = new self();
        }

        return self::$instance;
    }

    /**
     * @param mixed $debugMode
     * @param mixed $division_id
     * @param mixed $country
     */
    public function updateSite($division_id, $country, $debugMode = 0): array|\WP_Error
    {
        $url = $this->getEndpointUrl('update_site');
        $site_name = get_bloginfo('name', 'display');

        if (empty($site_name)) {
            // use site url to extract site name
            $site_name = parse_url(get_bloginfo('url'), PHP_URL_HOST);
        }

        $data = [
            'token' => 'bypass',
            'division_id' => (int) $division_id,
            'country' => strtolower($country),
            'name' => $site_name,
            'url' => get_bloginfo('url'),
            'version' => JIMS_CBP_VERSION,
            'test_mode' => $debugMode,
            'php_version' => phpversion(),
            'platform_name' => 'wordPress',
            'platform_version' => get_bloginfo('version'),
        ];

        return $this->postRequest($url, $data);
    }

    public function updateWebsite(): array|\WP_Error
    {
        $division_id = OptionHelper::getFormOption('bookingFormDivisionId');
        $country = OptionHelper::getFormOption('country');
        $debug_mode = OptionHelper::getFormOption('debugMode', 0);

        if (!empty($division_id)) {
            return $this->updateSite($division_id, $country, $debug_mode);
        }

        return new \WP_Error('missing_data', 'Division ID is missing');
    }

    public function getSiteStatus(): bool
    {
        $url = $this->getEndpointUrl('status');

        $request = $this->getRequest($url);

        if (is_wp_error($request)) {
            return false;
        }

        $response = $request['response'] ?? [];

        if (isset($response['code']) && $response['code'] === 200) {
            return true;
        }

        return false;
    }

    /**
     * @param null|mixed $error_code
     * @param mixed      $booking_id
     *
     * @throws \JsonException
     */
    public function updateRealTimeBooking($booking_id, $error_code = null): array|\WP_Error
    {
        $debug_mode = OptionHelper::getFormOption('debugMode', 0);

        if ($debug_mode == 1) {
            return [];
        }

        $url = $this->getEndpointUrl('update_realtime');
        $booking = new Booking();
        $booking_data = $booking->getFromId($booking_id, Booking::RESULT_TYPE_ARRAY_A);

        if (!empty($error_code)) {
            $booking_data['error_code'] = $error_code;
        }

        $realtime_booking = $this->formatRealTimeBooking($booking_data);

        if (
            empty($realtime_booking['suburb'])
            || empty($realtime_booking['state_code'])
            || empty($realtime_booking['country_code'])
        ) {
            error_log('Realtime Booking: Missing suburb, state code or country code', 0);

            return new \WP_Error(
                'missing_data',
                __FUNCTION__ . ' Missing suburb, state code or country code',
                $realtime_booking
            );
        }

        // add website url
        $realtime_booking['website'] = get_bloginfo('url');

        $response = Http::post($url, $realtime_booking, [
            'Content-Type: application/json',
            'Accept: application/json',
            'X-Auth-Token: wpcb',
        ], OptionHelper::isHTTPCompatibilityEnabled('internal'));

        return StringHelper::isJson($response['result'])
            ? json_decode($response['result'], true, 512, JSON_THROW_ON_ERROR)
            : [];
    }

    /**
     * @param null|mixed $error_code
     * @param mixed      $booking_id
     */
    public function sendErrorNotificationEmail($booking_id, $error_code = null): array|\WP_Error
    {
        $debug_mode = OptionHelper::getFormOption('debugMode', 0);

        if ($debug_mode == 1) {
            return [];
        }

        $url = $this->getEndpointUrl('email_failed');
        $booking = new Booking();
        $booking_data = $booking->getFromId($booking_id, Booking::RESULT_TYPE_ARRAY_A);

        // add token bypass to $booking_data
        $booking_data['token'] = 'bypass';

        return $this->sendRequest($url, $booking_data);
    }

    public function setMode(string $mode): void
    {
        if (!in_array($mode, $this->allowed_modes, true)) {
            throw new \InvalidArgumentException('Invalid mode');
        }

        $this->mode = $mode;
    }

    public function getStateCodeByName($name, $country = 'AU'): string
    {
        if (strlen($name) <= 3) {
            return strtoupper($name);
        }

        $states = $this->getStates($country);

        // flip array key and value
        $states = array_flip($states);

        return $states[$name] ?? '';
    }

    public function getStates($country = 'AU'): array
    {
        $states = [
            'AU' => [
                'ACT' => 'Australian Capital Territory',
                'NSW' => 'New South Wales',
                'NT' => 'Northern Territory',
                'QLD' => 'Queensland',
                'SA' => 'South Australia',
                'TAS' => 'Tasmania',
                'VIC' => 'Victoria',
                'WA' => 'Western Australia',
            ],
            'NZ' => [
                'N' => 'North Island',
                'S' => 'South Island',
            ],
            'CA' => [
                'AB' => 'Alberta',
                'BC' => 'British Columbia',
                'MB' => 'Manitoba',
                'NB' => 'New Brunswick',
                'NL' => 'Newfoundland and Labrador',
                'NS' => 'Nova Scotia',
                'NT' => 'Northwest Territories',
                'NU' => 'Nunavut',
                'ON' => 'Ontario',
                'PE' => 'Prince Edward Island',
                'QC' => 'Quebec',
                'SK' => 'Saskatchewan',
                'YT' => 'Yukon',
            ],
            'UK' => [
                'ENG' => 'England',
                'NIR' => 'Northern Ireland',
                'SCT' => 'Scotland',
                'WLS' => 'Wales',
            ],
            'GB' => [
                'ENG' => 'England',
                'NIR' => 'Northern Ireland',
                'SCT' => 'Scotland',
                'WLS' => 'Wales',
            ],
        ];

        return $states[$country] ?? [];
    }

    /**
     * @throws \JsonException
     */
    public function formatRealTimeBooking(array $booking_data): array
    {
        return [
            'division_id' => ArrayHelper::getProperty($booking_data, 'data.division_id'),
            'suburb' => ArrayHelper::getProperty($booking_data, 'data.address.suburb'),
            'state_code' => $this->getStateCodeByName(ArrayHelper::getProperty($booking_data, 'data.address.state')),
            'country_code' => ArrayHelper::getProperty($booking_data, 'data.address.country'),
            'services' => ArrayHelper::getProperty($booking_data, 'data.services'),
            'data' => $booking_data,
            'created_at' => ArrayHelper::getProperty($booking_data, 'created_at'),
        ];
    }

    private function getEndpointUrl($name): string
    {
        return $this->servers[$this->mode] . $this->endpoints[$this->mode][$name];
    }

    /**
     * Like sendRequest but using curl.
     */
    private function postRequest(string $url, array $data): array
    {
        return Http::post($url, $data, [
            'Content-Type: application/json',
            'Accept: application/json',
            'X-Auth-Token: wpcb',
        ], OptionHelper::isHTTPCompatibilityEnabled('internal'));
    }

    private function sendRequest($url, $data): array|\WP_Error
    {
        //        error_log(json_encode($data), 0);
        return \wp_remote_post($url, [
            'body' => $data,
            'headers' => [
                'Content-Type' => 'application/json',
                'Accept' => 'application/json',
                'X-Auth-Token' => 'wpcb',
            ],
        ]);
    }

    private function getRequest($url, array $params = []): array|\WP_Error
    {
        if (count($params) > 0) {
            $params_query = http_build_query($params);
            $url .= '?' . $params_query;
        }

        return \wp_remote_get($url, [
            'headers' => [
                'Content-Type' => 'application/json',
                'Accept' => 'application/json',
                'X-Auth-Token' => 'wpcb',
            ],
        ]);
    }
}
