<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\Database\Booking;
use JimsCustomerBooking\Helpers\TimeHelper;

class BookingsStatisticsController extends RestController
{
    protected string $route = 'booking-statistics';

    public function registerRoutes(): void
    {
        register_rest_route($this->getNamespace(), '/' . $this->getRoute() . '/totals', [[
            'methods' => \WP_REST_Server::READABLE,
            'callback' => [$this, 'getTotals'],
            'permission_callback' => '__return_true',
            'args' => $this->get_collection_params(),
        ],
        ]);

        register_rest_route($this->getNamespace(), '/' . $this->getRoute() . '/totals-by-date', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'getTotalsByDate'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params(),
            ],
        ]);

        register_rest_route($this->getNamespace(), '/' . $this->getRoute() . '/all', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'getAll'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params(),
            ],
        ]);

        register_rest_route($this->getNamespace(), '/' . $this->getRoute() . '/all/period', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'getAllByPeriod'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params(),
            ],
        ]);

        register_rest_route($this->getNamespace(), '/' . $this->getRoute() . '/period/total', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'getPeriodTotal'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params(),
            ],
        ]);
    }

    /**
     * @return \WP_REST_Response
     */
    public function getAll(\WP_REST_Request $request)
    {
        $page = $request->get_param('page');
        $per_page = $this->getPageLimitParam($request);

        if ($page == null) {
            $page = 1;
        }

        if ($per_page == null) {
            $per_page = 10;
        }

        $offset = ($page - 1) * $per_page;

        $result = array_map('self::transformSingle', Booking::getStatisticsAll($offset, $per_page));

        return $this->response([
            'status' => 'success',
            'page_count' => (int) ceil(Booking::getStatisticsAllPageCount() / $per_page),
            'current_page' => $page,
            'data' => $result,
        ]);
    }

    public function getAllByPeriod(\WP_REST_Request $request)
    {
        $status = $request->get_param('status');
        $start_date = $request->get_param('start_date');
        $end_date = $request->get_param('end_date');
        $page = $request->get_param('page');
        $per_page = $this->getPageLimitParam($request);
        $timezone = $this->getTimeZone($request);

        if ($page == null) {
            $page = 1;
        }

        if ($per_page == null) {
            $per_page = 10;
        }

        $offset = ($page - 1) * $per_page;

        $result = array_map(
            'self::transformSingle',
            Booking::getStatisticsAllByPeriod($status, $start_date, $end_date, $offset, $per_page, $timezone)
        );

        return $this->response([
            'status' => 'success',
            'page_count' => Booking::getPageCountForPeriod($start_date, $end_date, $per_page, $status),
            'current_page' => $page,
            'data' => $result,
        ]);
    }

    public function getPeriodTotal(\WP_REST_Request $request)
    {
        $status = $request->get_param('status');
        $start_date = $request->get_param('start_date');
        $end_date = $request->get_param('end_date');
        $timezone = $this->getTimeZone($request);

        return $this->response(Booking::getStatisticsPeriodTotal($status, $start_date, $end_date, $timezone));
    }

    /**
     * @return \WP_REST_Response
     */
    public function getTotals(\WP_REST_Request $request)
    {
        $status = $request->get_param('status');

        $start_date = $request->get_param('start_date');
        $end_date = $request->get_param('end_date');
        $timezone = $this->getTimeZone($request);

        return $this->response(Booking::getStatisticsTotals($status, $start_date, $end_date, $timezone));
    }

    /**
     * @return \WP_REST_Response
     */
    public function getTotalsByDate(\WP_REST_Request $request)
    {
        $status = $request->get_param('status');
        $date = $request->get_param('date');
        $timezone = $this->getTimeZone($request);

        return $this->response(Booking::getStatisticsTotalsByDate($date, $status, $timezone));
    }

    /**
     * Transforms a single booking from database format to API format.
     *
     * @param mixed $booking
     *
     * @return array
     */
    public static function transformSingle($booking)
    {
        $data = json_decode($booking['data'], true);

        return [
            'id' => $booking['id'],
            'title' => $booking['title'],
            'division_id' => !empty($booking['division_id'])
                ? $booking['division_id']
                : self::getArrayValue($data, 'division_id'),
            'address' => self::getArrayValue($data, 'address.full_address'),
            'mode' => self::getArrayValue($booking, 'mode'),
            'channel' => self::getArrayValue($booking, 'channel', 'online', true),
            'state' => self::getArrayValue($data, 'address.state'),
            'suburb' => self::getArrayValue($data, 'address.suburb'),
            'country' => self::getArrayValue($data, 'address.country'),
            'postcode' => self::getArrayValue($data, 'address.postcode'),
            'services' => self::getArrayValue($data, 'services'),
            'status' => $booking['status'],
            'error_code' => isset($booking['error_code']) ? $booking['error_code'] : null,
            'created_at' => date('d/m/Y H:i:s', strtotime($booking['created_at'])),
        ];
    }

    public static function getArrayValue($array, $key, $default = null, $return_default_if_empty = false)
    {
        // if key contains dot, get the value from the nested array
        if (str_contains($key, '.')) {
            $keys = explode('.', $key);
            $value = $array;

            foreach ($keys as $a_key) {
                $value = self::getArrayValue($value, $a_key);
            }

            if ($return_default_if_empty && empty($value)) {
                return $default;
            }

            return $value;
        }

        return isset($array[$key]) ? $array[$key] : $default;
    }

    public function getPageLimitParam(\WP_REST_Request $request): mixed
    {
        $per_page = $request->get_param('per_page');
        $page_limit = $request->get_param('page_limit');

        if ($page_limit !== null) {
            $per_page = $page_limit;
        }

        return $per_page;
    }

    public function convertDates(\WP_REST_Request $request, mixed $start_date, mixed $end_date): array
    {
        if ($request->has_param('timezone')) {
            $timezone = $request->get_param('timezone');

            // replace - with /
            $timezone = str_replace('-', '/', $timezone);

            // check if timezone has more than 3 characters
            if (strlen($timezone) > 3) {
                $start_date = TimeHelper::convertToUTC($start_date, $timezone);
                $end_date = TimeHelper::convertToUTC($end_date, $timezone);
            } else {
                $start_date = TimeHelper::convertDateToTimezoneFromCountryCode($start_date, $timezone);
                $end_date = TimeHelper::convertDateToTimezoneFromCountryCode($end_date, $timezone);
            }
        }

        return [$start_date, $end_date];
    }

    public function getTimezoneString(string $timezone): string
    {
        return strlen($timezone) > 3
            ? $timezone
            : TimeHelper::getTimezoneFromCountryCode($timezone);
    }

    public function getTimeZone(\WP_REST_Request $request): string
    {
        $timezone = '+00:00';

        if ($request->has_param('timezone')) {
            $timezone = $this->getTimezoneString($request->get_param('timezone'));
        }

        return $timezone;
    }
}
