<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\JimsPlugin;
use JimsCustomerBooking\Services\TokenManager;

if (!class_exists('JimsCustomerBooking\RestfulPI\RestController')) {
    /**
     * Class RestController.
     *
     * Instructions how to add a restful controller using this class
     * 1. Create a class that extends this class
     * 2. Add a route to the class
     * 3. Add a method to the class
     */
    abstract class RestController extends \WP_REST_Controller
    {
        /**
         * Plugin's API namespace.
         *
         * @var string
         */
        protected $namespace = 'jims-wpcb';

        /**
         * Plugin's API version.
         *
         * @var string
         */
        protected $version = '1';

        /**
         * Plugin's API default route.
         */
        protected string $route = 'default';

        protected string $key = JimsPlugin::SECKEY;

        public function __construct()
        {
            $this->registerRoutes();
        }

        abstract public function registerRoutes(): void;

        /**
         * @param $request \WP_REST_Request
         *
         * @return bool
         */
        public function checkPermissions($request)
        {
            $allowed_tokens = [
                'bypass', // bypass token
            ];

            if ($request->has_param('token')) {
                $token = $request->get_param('token');

                if (in_array($token, $allowed_tokens)) {
                    return true;
                }

                if (TokenManager::verify($this->key, $token)) {
                    return true;
                }

                // verify token
                if (wp_verify_nonce($request->get_param('token'), 'wp_rest')) {
                    return true;
                }
            }

            // if request has no token, return false
            return false;
        }

        /**
         * @return string
         */
        protected function getNamespace()
        {
            return "{$this->namespace}/v{$this->version}";
        }

        protected function getRoute(string $name = ''): string
        {
            return '/' . ($name !== '' ? $name : $this->route);
        }

        /**
         * @param int   $http_code
         * @param mixed $data
         *
         * @return \WP_REST_Response
         */
        protected function response($data, $http_code = 200)
        {
            return new \WP_REST_Response($data, $http_code);
        }

        /**
         * @return \WP_REST_Response
         */
        protected function emptyResponse()
        {
            return $this->response([]);
        }

        /**
         * @param mixed      $param_name
         * @param null|mixed $custom_message
         *
         * @return array
         */
        protected function invalidRequest($param_name = false, $custom_message = null)
        {
            $message = 'Invalid Request' . ($param_name == false ? '' : ": Missing or invalid parameter {$param_name}");
            if (is_null($custom_message)) {
                return [
                    'status' => 404,
                    'message' => $message,
                ];
            }

            return [
                'status' => 404,
                'message' => $custom_message,
            ];
        }

        /**
         * @param null|mixed $custom_message
         *
         * @return array
         */
        protected function successResponse($custom_message = null)
        {
            return [
                'status' => 200,
                'message' => 'Your action has successful',
            ];
        }

        /**
         * @param null|mixed $custom_message
         *
         * @return array
         */
        protected function failedResponse($custom_message = null)
        {
            return [
                'status' => 500,
                'message' => 'Your action failed',
            ];
        }

        protected function getWordpressRestURL(string $path = ''): string
        {
            return get_rest_url(null, $this->getNamespace() . $path);
        }

        protected function transientExists($transient_name)
        {
            return get_transient($transient_name) !== false;
        }

        protected function storedResponseIsEmpty($response)
        {
            // check if response is empty or the length of the response is less than 2
            return empty($response) || (is_string($response) && strlen($response) < 2);
        }

        protected function getRequestParam($request, $param_name, $default = null)
        {
            return $request->get_param($param_name) ?? $default;
        }

        protected function registerRestRoute($namespace, $routes, $args = [], $override = false)
        {
            if (!empty($routes)) {
                foreach ($routes as $route) {
                    register_rest_route($namespace, $route, $args, $override);
                }
            }
        }
    }
}
