<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Services\DivisionService;

if (!class_exists('JimsCustomerBooking\RestfulPI\ServiceController')) {
    class ServiceController extends RestController
    {
        protected string $route = 'service';

        public function registerRoutes(): void
        {
            $this->registerRestRoute($this->getNamespace(), [
                $this->getRoute() . '/divisions',
                $this->getRoute() . '/division',
            ], [
                [
                    'methods' => \WP_REST_Server::READABLE,
                    'callback' => [$this, 'getDivisions'],
                    'permission_callback' => '__return_true',
                ],
            ]);

            $this->registerRestRoute($this->getNamespace(), [$this->getRoute() . '/division/services'], [
                [
                    'methods' => \WP_REST_Server::READABLE,
                    'callback' => [$this, 'getServices'],
                    'permission_callback' => '__return_true',
                ],
            ]);

            $this->registerRestRoute($this->getNamespace(), [$this->getRoute() . '/division/icons'], [
                [
                    'methods' => \WP_REST_Server::READABLE,
                    'callback' => [$this, 'getIcons'],
                    'permission_callback' => '__return_true',
                ],
            ]);

            $this->registerRestRoute($this->getNamespace(), [$this->getRoute() . '/division/options'], [
                [
                    'methods' => \WP_REST_Server::READABLE,
                    'callback' => [$this, 'getDivisionOptions'],
                    'args' => [
                        'division_id' => [
                            'required' => true,
                            'type' => 'integer',
                        ],
                    ],
                    'permission_callback' => '__return_true',
                ],
            ]);

            $this->registerRestRoute($this->getNamespace(), [$this->getRoute() . '/countries'], [
                [
                    'methods' => \WP_REST_Server::READABLE,
                    'callback' => [$this, 'getCountries'],
                    'permission_callback' => '__return_true',
                ],
            ]);
        }

        /**
         * @param mixed $request
         *
         * @return \WP_REST_Response
         *
         * @throws \Exception
         */
        public function getDivisions($request)
        {
            $service = new DivisionService();
            $countryCode = $this->getRequestParam($request, 'country_code', 'AU');

            return $this->response($service->getDivisions(true, $countryCode));
        }

        public function getDivisionOptions(\WP_REST_Request $request): \WP_REST_Response
        {
            $division_id = $request->get_param('division_id');

            if ((int) $division_id === 0) {
                return $this->response([]);
            }

            $customize_options = OptionHelper::get(
                'division_' . $division_id . '_customer_booking_customizer_options',
                []
            );

            if (count($customize_options) === 0) {
                $customize_options = OptionHelper::get('customer_booking_customizer_options', []);
            }

            if (count($customize_options) === 0) {
                return $this->response([
                    'alphabetical_order' => false,
                    'custom_order' => false,
                    'categories_enabled' => false,
                    'service_descriptions_enabled' => false,
                ]);
            }

            return $this->response($customize_options);
        }

        /**
         * @return \WP_REST_Response
         */
        public function getServices(\WP_REST_Request $request)
        {
            $countryCode = $this->getRequestParam($request, 'country_code', 'AU');
            $serviceListName = $this->getRequestParam($request, 'service_list', '');
            $service = new DivisionService();

            if ($request->has_param('division_id') && (int) $request->get_param('division_id') > 0) {
                try {
                    $payload = $service->getServices(false, $countryCode, $serviceListName);
                    $division_id = $request->get_param('division_id');

                    $customize_options = OptionHelper::get(
                        'division_' . $division_id . '_customer_booking_customizer_options',
                        []
                    );

                    if (count($customize_options) === 0) {
                        $customize_options = OptionHelper::get('customer_booking_customizer_options', []);
                    }

                    $filtered = array_filter($payload, function ($service) use ($request) {
                        return $service['group_id'] == $request->get_param('division_id');
                    });

                    $filtered = array_filter($filtered, function ($service) {
                        return !empty(trim($service['friendly_name']));
                    });



                    $use_grouping = $request->get_param('use_grouping');
                    if ($use_grouping === '1') {
                        $filtered = $this->groupServices($filtered, $customize_options, $division_id);
                    }

                    if ($serviceListName === '') {
                        // disable custom order if custom service list is enabled
                        // if custom order is enabled
                        $custom_order_enabled = (
                            isset($customize_options['custom_order'])
                            && $customize_options['custom_order']
                        ) == true;

                        if ($custom_order_enabled) {
                            $custom_order = OptionHelper::get(
                                'division_' . $division_id . '_customer_booking_service_order',
                                []
                            );

                            if (count($custom_order) === 0) {
                                $custom_order = OptionHelper::get('customer_booking_service_order', []);
                            }

                            $filtered = $this->orderServices($custom_order, $filtered, !empty($use_grouping));
                        } else {
                            // use fms order by default
                            usort($filtered, function ($a, $b) {
                                return ($a['online_order'] ?? 0) > ($b['online_order'] ?? 0) ? 1 : -1;
                            });
                        }
                    }
                } catch (\Exception $e) {
                    return $this->response([]);
                }

                return $this->response($filtered);
            }

            return $this->response($service->getServices(false, $countryCode));
        }

        /**
         * @param mixed $request
         *
         * @return \WP_REST_Response
         *
         * @throws \JsonException
         */
        public function getCountries($request)
        {
            $service = new DivisionService();

            return $this->response($service->getCountries());
        }

        /**
         * @param mixed $division_id
         * @param mixed $filtered
         * @param mixed $customize_options
         *
         * @return array
         */
        public function groupServices($filtered, $customize_options, $division_id)
        {
            $groups = OptionHelper::get('division_' . $division_id . '_customer_booking_service_groups', []);

            if (count($groups) === 0) {
                $groups = OptionHelper::get('customer_booking_service_groups', []);
            }

            // get if service description is enabled
            $service_description_enabled = isset($customize_options['service_descriptions_enabled'])
                && $customize_options['service_descriptions_enabled'] === true;

            $filtered = $this->addDescriptions($service_description_enabled, $filtered, $division_id);

            // get groups
            $grouped = [];
            $grouped_ids = [];
            foreach ($groups as $group) {
                if (!empty($group['services'])) {
                    $group['services'] = $this->getGroupServices($filtered, $group['services']);
                    $grouped[] = $group;
                }
            }

            // remove grouped services from filtered
            $filtered = array_filter($filtered, function ($service) use ($grouped_ids) {
                return !in_array($service['id'], $grouped_ids);
            });

            // create a custom group for the remaining services
            if (!empty($filtered)) {
                $grouped[] = $this->createRestGroup($filtered);
            }

            return $grouped;
        }

        /**
         * @param mixed $services
         *
         * @return array|null[]
         */
        public function getGroupServices(mixed &$filtered, $services): array
        {
            return array_map(function ($service) use (&$filtered) {
                $service = array_filter($filtered, function ($filtered_service) use ($service) {
                    return $filtered_service['id'] == $service['id'];
                });

                $service = array_shift($service);
                $grouped_ids[] = $service['id'];

                // remove service from filtered
                $filtered = array_filter($filtered, function ($filtered_service) use ($service) {
                    return $filtered_service['id'] != $service['id'];
                });

                return $service;
            }, $services);
        }

        /**
         * @param $request \WP_REST_Request
         */
        public function getIcons(\WP_REST_Request $request): \WP_REST_Response
        {
            $service = new DivisionService();

            return $this->response($service->getDivisionIcons());
        }

        /**
         * @param mixed $name
         * @param mixed $description
         */
        public function createRestGroup(mixed $filtered, $name = 'Other', $description = ''): array
        {
            return [
                'order' => 0,
                'description' => $description,
                'ignore_order' => 1,
                'name' => $name,
                'services' => $filtered,
            ];
        }

        /**
         * @param mixed $services_order
         * @param mixed $filtered
         *
         * @return array
         */
        protected function orderServices($services_order, $filtered, bool $is_grouped)
        {
            if ($is_grouped) {
                foreach ($filtered as $index => $group) {
                    usort($group['services'], function ($a, $b) use ($services_order) {
                        $a_index = array_search($a['id'], array_column($services_order, 'id'));
                        $b_index = array_search($b['id'], array_column($services_order, 'id'));

                        return ($services_order[$a_index]['gravity'] > $services_order[$b_index]['gravity']) ? -1 : 1;
                    });

                    $filtered[$index]['services'] = $group['services'];

                    // sum the gravity of the services in the group
                    $gravity_sum = array_sum(array_map(function ($service) use ($services_order) {
                        $index = array_search($service['id'], array_column($services_order, 'id'));

                        return $services_order[$index]['gravity'];
                    }, $group['services']));

                    if (!(isset($filtered[$index]['ignore_order']) && $filtered[$index]['ignore_order'] === 1)) {
                        $filtered[$index]['order'] = $gravity_sum / count($group['services']);
                    }
                }
            }

            usort($filtered, function ($a, $b) use ($services_order) {
                if (array_key_exists('id', $a) && array_key_exists('id', $b)) {
                    $a_index = array_search($a['id'], array_column($services_order, 'id'));
                    $b_index = array_search($b['id'], array_column($services_order, 'id'));

                    return ($services_order[$a_index]['gravity'] > $services_order[$b_index]['gravity']) ? -1 : 1;
                }

                return 0;
            });

            return $filtered;
        }

        protected function getDescription($service, $service_descriptions)
        {
            $descriptions = array_values(array_filter($service_descriptions, function ($description) use ($service) {
                return $description['id'] == $service['id'];
            }));

            return array_shift($descriptions);
        }

        /**
         * @param mixed $division_id
         * @param mixed $filtered
         */
        protected function addDescriptions(bool $service_description_enabled, $filtered, $division_id): mixed
        {
            if ($service_description_enabled) {
                // get descriptions
                $service_descriptions = OptionHelper::get(
                    'division_' . $division_id . '_customer_booking_service_descriptions',
                    []
                );

                if (count($service_descriptions) === 0) {
                    $service_descriptions = OptionHelper::get('customer_booking_service_descriptions', []);
                }

                // populate filtered with descriptions
                $filtered = array_map(function ($service) use ($service_descriptions) {
                    // find description
                    $description = $this->getDescription($service, $service_descriptions);
                    $service['description'] = $description['description'] ?? '';

                    return $service;
                }, $filtered);
            }

            return $filtered;
        }
    }
}
