<?php

namespace JimsGroup\Apis;

use JimsGroup\Utilities\HttpHeaders;
use JimsGroup\Utilities\HttpRequest;

//use JimsGroup\Utilities\Logger;

/**
 * Class AbstractApi
 * @package JimsGroup\Apis
 */
abstract class AbstractApi
{
    /**
     * Refers to the api urls keys
     * @var string
     */
    protected string $mode = 'production';


    protected $compatibility_mode = false;


    /**
     * @var string[] $api_urls
     */
    protected $api_urls = [
        'production' => 'https://api.jims.net',
        'uat' => 'https://api.uat.aws.jims.net', //user acceptance testing
        'hf' => 'https://api.hf.aws.jims.net', // hotfix
        'staging' => 'https://api.staging.aws.jims.net', // staging, it can create real bookings
        'int' => 'https://api.int.aws.jims.net/', //integration
        'development' => 'https://api.int.aws.jims.net',
        'local' => 'http://localhost:3000',
    ];

    private $response_headers = [];

    /**
     * @var array $paths
     */
    protected $paths = [];

    public function __construct($mode = 'production')
    {
        $this->mode = $mode;
    }

    /**
     * @param string $name
     * @return string
     */
    public function getEndpointUrl(string $name): string
    {
        return $this->getApiUrl() . $this->paths[$name];
    }

    /**
     * @return string
     */
    public function getApiUrl(): string
    {
        if (isset($this->api_urls[$this->mode])) {
            return $this->api_urls[$this->mode];
        }

        throw new \Exception('Invalid mode: ' . $this->mode . ' for ' . get_class($this) . ' api.');
    }

    /**
     * @return string[] $api_urls
     */
    public function getApiUrls(): array
    {
        return $this->api_urls;
    }

    /**
     * @param mixed|string $mode
     */
    public function setMode(mixed $mode): void
    {
        $this->mode = $mode;
    }


    /**
     * @param $url string
     * @param $params array
     * @return string
     * @throws \JsonException
     */
    protected function post(
        string $url,
        array $params = [],
        array $extra_headers = [],
        $use_compatibility_mode = false
    ): string {
        $request = new HttpRequest();
        $request->setUrl($url);
        $request->setMethod('POST');
        $request->setHeaders([
            'Content-Type: application/json',
            'Accept: application/json',
        ]);

        if ($this->compatibility_mode) {
            $request->enableCompatibilityMode();
        }

        if (count($extra_headers) > 0) {
            foreach ($extra_headers as $header) {
                $request->addHeader($header);
            }
        }

        $request->setBody(json_encode($params, JSON_THROW_ON_ERROR));
        $request->send();

        $this->response_headers = $request->getResponseHeaders(true);

        return $request->getResponse();
    }

    public function put(
        string $url,
        array $params = [],
        array $extra_headers = [],
        $use_compatibility_mode = false
    ): string {
        $request = new HttpRequest();
        $request->setUrl($url);
        $request->setMethod('PUT');
        $request->setHeaders([
            'Content-Type: application/json',
            'Accept: application/json',
        ]);

        if ($this->compatibility_mode) {
            $request->enableCompatibilityMode();
        }

        if (count($extra_headers) > 0) {
            foreach ($extra_headers as $header) {
                $request->addHeader($header);
            }
        }

        $request->setBody(json_encode($params, JSON_THROW_ON_ERROR));
        $request->send();

        $this->response_headers = $request->getResponseHeaders(true);

        return $request->getResponse();
    }

    /**
     * @param $url
     * @param array $params
     * @param array $extra_headers
     * @return string
     */
    protected function get(
        string $url,
        array $params = [],
        array $extra_headers = [],
        $use_compatibility_mode = false
    ): string {
        $request_headers = new HttpHeaders([
            'Content-Type' =>  'application/json',
            'Accept' => 'application/json',
        ]);

        if (count($extra_headers) > 0) {
            foreach ($extra_headers as $header_name => $header) {
                $request_headers->add($header_name, $header);
            }
        }

        if (count($params) > 0 && !empty($params)) {
            $params_query = http_build_query($params);
            $url = $url . '?' . $params_query;
        }

        $request = new HttpRequest();
        if ($this->compatibility_mode) {
            $request->enableCompatibilityMode();
        }
        $request->setUrl($url);
        $request->setMethod('GET');
        $request->setHeaders($request_headers->toCurlArray());
        $request->send();

        //uncomment to log headers
//        $headers = $request->getResponseHeaders(true);
//        error_log(json_encode($headers), 0);

        return $request->getResponse();
    }

    /**
     * @return bool
     */
    public function isCompatibilityMode(): bool
    {
        return $this->compatibility_mode;
    }

    /**
     * Compatibility mode is used to set the curl options to be compatible with older server configurations
     * @param bool $compatibility_mode
     */
    public function setCompatibilityMode(bool $compatibility_mode): void
    {
        $this->compatibility_mode = $compatibility_mode;
    }

    public function isJSON(string $data)
    {
        json_decode($data);
        return (json_last_error() == JSON_ERROR_NONE);
    }
}
