<?php

namespace JimsGroup\Apis;

use JimsGroup\Apis\AbstractApi;
use JimsGroup\Datatypes\DynamicPropertiesTrait;
use JimsGroup\Datatypes\Franchisee\AddressValidationResponseDataType;
use JimsGroup\Datatypes\Franchisee\AllocationRequestDataType;
use JimsGroup\Datatypes\Franchisee\AllocationResponseDataType;
use JimsGroup\Datatypes\Franchisee\ClientDataType;
use JimsGroup\Datatypes\Franchisee\EnquiryResponseDataType;
use JimsGroup\Datatypes\Franchisee\ProspectRequestDataType;
use JimsGroup\Datatypes\Franchisee\ProspectResponseDataType;
use JimsGroup\Utilities\Logger;
use PHPUnit\TextUI\RuntimeException;

class FranchiseeEnquiry extends AbstractApi
{
    use DynamicPropertiesTrait;

    protected string $mode = 'production';

    protected bool $tryToParseJson = false;

    /**
     * @var string[]
     */
    protected $paths = [
        'enquiry' => '/api/bookings/new-enquiry',
        'allocation' => '/api/bookings/allocation-request',
        'validate' => '/api/bookings/validation-address/{{booking_id}}',
        'request' => '/api/bookings/send-request',
    ];

    private $requests = [];



    /**
     * @param $divisionId
     * @param $accessToken
     * @return string|EnquiryResponseDataType
     * @throws \JsonException|RuntimeException
     */
    public function prepareEnquiry($divisionId, $accessToken): string|EnquiryResponseDataType
    {
        $url = $this->getEndpointUrl('enquiry');

        //create payload
        Logger::info('Sending request to ' . $url, JSON_THROW_ON_ERROR);
        $response = $this->get($url, [
            'divisionId' => $divisionId,
        ], [
            'Authorization' => 'Bearer  ' . $accessToken,
        ]);

        if ($this->tryToParseJson && $this->isJSON($response)) {
            $data = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
            return new EnquiryResponseDataType($data);
        }

        return $response;
    }

    /**
     * @throws \JsonException|RuntimeException
     */
    public function validateAddress(
        $booking_id,
        $addressPartial,
        $accessToken
    ): string|AddressValidationResponseDataType {
        $url = $this->getEndpointUrl('validate', ['booking_id' => $booking_id]);

        //create payload
        Logger::info('Sending request to ' . $url, JSON_THROW_ON_ERROR);
        $response = $this->put($url, [
            'requestedAddress' => $addressPartial,
            'acceptNonCompletedAddress' => true,
        ], [
            'Authorization: Bearer ' .  $accessToken,
        ]);

        if ($this->tryToParseJson && $this->isJSON($response)) {
            $data = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
            return new AddressValidationResponseDataType($data);
        }

        return $response;
    }

    /**
     * @throws \JsonException|RuntimeException
     */
    public function sendRequest(
        $booking_id,
        ClientDataType $client,
        $selectedBusinessId,
        $accessToken
    ): string|ProspectResponseDataType {
        $url = $this->getEndpointUrl('request');
        $prospectData = new ProspectRequestDataType(
            array_merge($client->toArray(), ['booking_id' => $booking_id, 'selectedBusinessId' => $selectedBusinessId])
        );


        //create payload
        Logger::info('Sending request to ' . $url, JSON_THROW_ON_ERROR);
        $response = $this->post($url, $prospectData->toArray(), [
            'Authorization: Bearer ' . $accessToken,
        ]);

        if ($this->tryToParseJson && $this->isJSON($response)) {
            $data = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
            return new ProspectResponseDataType($data);
        }

        return $response;
    }

    /**
     *
     * @throws \JsonException|RuntimeException
     */
    public function allocate($booking_id, $accessToken): string|AllocationResponseDataType
    {
        $url = $this->getEndpointUrl('allocation');
        $allocationData = new AllocationRequestDataType(['booking_id' => $booking_id]);

        //create payload
        Logger::info('Sending request to ' . $url, JSON_THROW_ON_ERROR);
        $response = $this->post($url, $allocationData->toArray(), [
            'Authorization: Bearer ' . $accessToken,
        ]);

        if ($this->tryToParseJson && $this->isJSON($response)) {
            $data = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
            return new AllocationResponseDataType($data);
        }

        return $response;
    }

    public function getEndpointUrl(string $name, array $params = []): string
    {
        $path = $this->paths[$name];
        //replace placeholders in path
        if (count($params) > 0) {
            $placeholders = array_map(function ($key) {
                return '{{' . $key . '}}';
            }, array_keys($params));

            $path = str_replace($placeholders, array_values($params), $this->paths[$name]);
        }

        return $this->getApiUrl() . $path;
    }

    public function getMode(): string
    {
        return $this->mode;
    }

    public function setMode(mixed $mode): void
    {
        $this->mode = $mode;
    }

    public function getRequests(): array
    {
        return $this->requests;
    }

    public function setRequests(array $request, $name): void
    {
        $this->requests[$name] = $request;
    }

    public function setTryToParseJson(bool $tryToParseJson): void
    {
        $this->tryToParseJson = $tryToParseJson;
    }
}
