<?php

namespace JimsGroup\Utilities;

/**
 * the supported methods are GET, POST, PUT, DELETE, PATCH
 */
class HttpRequest
{
    protected int $query_encoding = \PHP_QUERY_RFC1738;

    public string $url = '';

    public string $method = 'GET';

    public array $headers = [];

    public string $body = '';

    public string $response = '';

    public int $responseCode = 0;

    public array|string $responseHeaders = [];

    public array $curl_opt_options = [];


    public function __construct(array $curl_opt_options = [])
    {
        if (count($curl_opt_options) > 0) {
            $this->curl_opt_options = $curl_opt_options;
        }
    }

    /**
     * @param string $url
     */
    public function setUrl($url)
    {
        $this->url = $url;
    }

    /**
     * @param string $method
     */
    public function setMethod($method)
    {
        $this->method = $method;
    }

    /**
     * @param array $headers
     */
    public function setHeaders($headers): void
    {
        $this->headers = $headers;
    }

    public function addHeader($header): void
    {
        $this->headers[] = $header;
    }

    /**
     * @param string $body
     */
    public function setBody($body): void
    {
        $this->body = $body;
    }

    /**
     * @return string
     */
    public function getResponse(): string
    {
        return $this->response;
    }

    /**
     * @return int
     */
    public function getResponseCode(): int
    {
        return $this->responseCode;
    }

    /**
     * @return array|string
     */
    public function getResponseHeaders($toArray = false)
    {
        return !!$toArray ? $this->headersToArray($this->responseHeaders) : $this->responseHeaders;
    }

    /**
     * @return string
     */
    public function send()
    {
        $ch = \curl_init();

        //set the default options
        $default_options = [
            CURLOPT_URL => $this->url,
            CURLOPT_HTTPHEADER => $this->headers,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_HEADER => true,
        ];

        switch ($this->method) {
            case 'POST':
                $default_options[CURLOPT_POST] = true;
                break;
            case 'PUT':
                $default_options[CURLOPT_CUSTOMREQUEST] = 'PUT';
                $default_options[CURLOPT_PUT] = true;
                break;
            case 'DELETE':
                $default_options[CURLOPT_CUSTOMREQUEST] = 'DELETE';
                break;
            case 'PATCH':
                $default_options[CURLOPT_CUSTOMREQUEST] = 'PATCH';
                break;
            default:
                $default_options[CURLOPT_HTTPGET] = true;
        }


        //if the method is not GET, set the custom request, otherwise, it will default to GET
        //this constant sometimes throws an error if is set to GET, since GET is the default
        if ($this->method !== 'GET') {
            $default_options[CURLOPT_POSTFIELDS] = $this->body;
        }

        //override or add the default options from $this->curl_opt_options, without using array_merge
        $options = $this->combineCurlOptions($default_options);


        Logger::info($options, 'cUrl options');

        //define the request
        curl_setopt_array($ch, $options);

        Logger::info($this->body);

        $time = microtime(true);
        $response = PerformanceTimer::ProductionTime(static function () use ($ch) {
            return \curl_exec($ch);
        }, $time);

        Logger::info("Request {$this->url}, duration: {$time} sec");
        Logger::info($this->response);

        //parse the response
        $header_size = \curl_getinfo($ch, CURLINFO_HEADER_SIZE);
        $this->responseHeaders = \substr($response, 0, $header_size);
        $this->response = \substr($response, $header_size);
        $this->responseCode = \curl_getinfo($ch, CURLINFO_HTTP_CODE);

        \curl_close($ch);
    }

    /**
     * @param string $url
     * @param array $params
     * @return string
     */
    public function get($url, array $params = [])
    {
        //if there are params, add them to the url
        if (count($params) > 0) {
            $url .= '?' . \http_build_query($params, '', '&', $this->query_encoding);
        }

        $this->setUrl($url);
        $this->setMethod('GET');
        $this->send();
        return $this->getResponse();
    }

    /**
     * @param string $url , the url to post to
     * @param array $params , the data to be encoded
     * @param string $enctype , how the data should be encoded
     * @return string
     */
    public function post($url, $params, $enctype = 'application/x-www-form-urlencoded')
    {

        $body = $this->formatBody($enctype, $params);

        $this->setUrl($url);
        $this->setMethod('POST');

        //set headers
        $this->addHeader('Content-Type: ' . $enctype);
        $this->addHeader('Content-Length: ' . strlen($body));

        $this->setBody($body);
        $this->send();

        return $this->getResponse();
    }

    /**
     * @param mixed $enctype , how the data should be encoded
     * @param array $params , the data to be encoded
     * @return string
     */
    protected function formatBody(mixed $enctype, array $params)
    {
        $body = '';
        if ($enctype == 'application/x-www-form-urlencoded') {
            $body = \http_build_query($params, '', '&', \PHP_QUERY_RFC1738);
        } elseif ($enctype == 'multipart/form-data') {
            $body = $params;
        } elseif ($enctype == 'application/json') {
            $body = \json_encode($params);
        }

        return $body;
    }

    /**
     * @param $jwt_token
     * @param $secret
     * @return array|bool
     */
    public static function parseJWT($jwt_token, $secret = '')
    {
        Logger::info($jwt_token);
        $jwt_token = explode('.', $jwt_token);
        $jwt_header = json_decode(base64_decode($jwt_token[0]), true);
        $jwt_payload = json_decode(base64_decode($jwt_token[1]), true);
        $jwt_signature = $jwt_token[2];

//        $signature = hash_hmac('sha256', $jwt_token[0] . '.' . $jwt_token[1], $secret, true);
//        $signature = base64_encode($signature);

//        return $signature === $jwt_signature ? $jwt_payload : false;
        return count($jwt_payload) > 0 ? $jwt_payload : false;
    }

    /**
     * @param $jwt_token
     * @return bool
     */
    public static function isJWT($jwt_token)
    {
        $jwt_token = explode('.', $jwt_token);
        return count($jwt_token) === 3;
    }

    /**
     * @param $payload
     * @param $secret
     * @return string
     */
    public static function createJWT($payload, $secret): string
    {
        $header = array(
            'alg' => 'HS256',
            'typ' => 'JWT'
        );

        $header = base64_encode(json_encode($header));
        $payload = base64_encode(json_encode($payload));
        $signature = hash_hmac('sha256', $header . '.' . $payload, $secret, true);
        $signature = base64_encode($signature);
        $jwt = $header . '.' . $payload . '.' . $signature;
        return $jwt;
    }

    /**
     * @param int $query_encoding
     */
    public function setQueryEncoding(int $query_encoding): void
    {
        $this->query_encoding = $query_encoding;
    }

    public function headersToArray($data)
    {
        $headers = [];
        if (is_array($data)) {
            return $data;
        }

        //extract the http status
        $status_arr = $this->getStatusFromHeadersString($data);

        $headers_arr = explode("\n", $data);
        foreach ($headers_arr as $header) {
            $header = explode(':', $header);
            if (count($header) == 2) {
                $headers[$header[0]] = trim($header[1]);
            }
        }
        return $status_arr + $headers;
    }

    /**
     * @param $data
     * @param $matches
     * @return mixed
     */
    protected function getStatusFromHeadersString($data): mixed
    {
        preg_match('/HTTP\/.* ([0-9]{3})/', $data, $matches);
        return count($matches) > 0 ? [
            'status' => $matches[1] ?? $matches[0],
        ] : [];
    }

    /**
     * @param string $version
     */
    public function enableCompatibilityMode($version = '1.1')
    {
        $versions = [
            '1.0' => CURL_HTTP_VERSION_1_0,
            '1.1' => CURL_HTTP_VERSION_1_1,
            '2.0' => CURL_HTTP_VERSION_2_0,
        ];

        if (isset($versions[$version])) {
            $this->setCurlOptOptions([
                CURLOPT_HTTP_VERSION => $versions[$version],
            ]);
        } else {
            $this->setCurlOptOptions([
                CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            ]);
        }
    }

    /**
     * @return array
     */
    public function getCurlOptOptions(): array
    {
        return $this->curl_opt_options;
    }

    /**
     * @param array $curl_opt_options
     * @param bool $reset
     * @return array
     */
    public function setCurlOptOptions(array $curl_opt_options, bool $reset = false): array
    {
        if (count($curl_opt_options) > 0) {
            $options = $reset || count($this->curl_opt_options) === 0
                ? $curl_opt_options
                : array_merge($this->curl_opt_options, $curl_opt_options);
            $this->curl_opt_options = $options;
        }

        return $this->curl_opt_options;
    }

    /**
     * @param array $default_options
     * @return array
     */
    public function combineCurlOptions(array $default_options): array
    {
        $options = $this->curl_opt_options;
        foreach ($options as $key => $value) {
            if (!isset($default_options[$key]) && $key !== CURLOPT_POSTFIELDS && $key !== CURLOPT_HTTPHEADER) {
                $default_options[$key] = $value;
            }
        }

        if (isset($options[CURLOPT_HTTPHEADER])) {
            $default_options[CURLOPT_HTTPHEADER] = array_merge(
                $default_options[CURLOPT_HTTPHEADER],
                $options[CURLOPT_HTTPHEADER]
            );
        }

        return $this->validateCurlOptions($default_options);
    }

    public function validateCurlOptions(array $options): array
    {
        $valid_options = [];
        foreach ($options as $key => $value) {
            if (is_numeric($key)) {
                $valid_options[(int)$key] = $value;
            }
        }

        return $valid_options;
    }
}
